# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest
import datetime
import logging

from stalker.db import DBSession
from stalker import (db, defaults, Studio, WorkingHours, Project, StatusList,
                     Status, Repository, Asset, Type, User, Shot, Department,
                     Task, TaskJugglerScheduler, SchedulerBase)


class DummyScheduler(SchedulerBase):
    """this is a dummy scheduler to be used in tests
    """

    def __init__(self, studio=None, callback=None):
        SchedulerBase.__init__(self, studio)
        self.callback = callback

    def schedule(self):
        """call the callback function before finishing
        """
        if self.callback:
            self.callback()


class StudioTester(unittest.TestCase):
    """tests the stalker.models.studio.Studio class
    """

    @classmethod
    def setUpClass(cls):
        """setup once
        """
        logger = logging.getLogger('stalker.models.schedulers')
        logger.setLevel(logging.DEBUG)

    def setUp(self):
        """setup the test
        """
        db.setup()
        db.init()

        self.test_user1 = User(
            name='User 1',
            login='user1',
            email='user1@users.com',
            password='password'
        )
        DBSession.add(self.test_user1)

        self.test_user2 = User(
            name='User 2',
            login='user2',
            email='user2@users.com',
            password='password'
        )
        DBSession.add(self.test_user2)

        self.test_user3 = User(
            name='User 3',
            login='user3',
            email='user3@users.com',
            password='password'
        )
        DBSession.add(self.test_user3)

        self.test_department1 = Department(
            name='Test Department 1'
        )
        DBSession.add(self.test_department1)

        self.test_department2 = Department(
            name='Test Department 2'
        )
        DBSession.add(self.test_department2)

        self.test_status1 = Status(
            name='Status 1',
            code='ST1'
        )
        DBSession.add(self.test_status1)

        self.test_status2 = Status(
            name='Status 2',
            code='ST2'
        )
        DBSession.add(self.test_status2)

        self.test_project_status_list1 = StatusList(
            name='Project Statuses',
            statuses=[self.test_status1, self.test_status2],
            target_entity_type=Project
        )
        DBSession.add(self.test_project_status_list1)

        self.test_repo = Repository(
            name='Test Repository',
            windows_path='T:/',
            linux_path='/mnt/T/',
            osx_path='/Volumes/T/'
        )
        DBSession.add(self.test_repo)

        # create a couple of projects
        self.test_project1 = Project(
            name='Test Project 1',
            code='TP1',
            status_list=self.test_project_status_list1,
            repository=self.test_repo
        )
        DBSession.add(self.test_project1)

        self.test_project2 = Project(
            name='Test Project 2',
            code='TP2',
            status_list=self.test_project_status_list1,
            repository=self.test_repo
        )
        DBSession.add(self.test_project2)

        # an inactive project
        self.test_project3 = Project(
            name='Test Project 3',
            code='TP3',
            status_list=self.test_project_status_list1,
            repository=self.test_repo
        )
        self.test_project3.active = False
        DBSession.add(self.test_project3)

        # create assets and shots
        self.test_asset_type = Type(
            name='Character',
            code='Char',
            target_entity_type=Asset
        )
        DBSession.add(self.test_asset_type)

        self.test_asset_status_list = \
            StatusList.query.filter_by(target_entity_type='Asset').first()

        self.test_asset1 = Asset(
            name='Test Asset 1',
            code='TA1',
            project=self.test_project1,
            type=self.test_asset_type
        )
        DBSession.add(self.test_asset1)

        self.test_asset2 = Asset(
            name='Test Asset 2',
            code='TA2',
            project=self.test_project2,
            type=self.test_asset_type
        )
        DBSession.add(self.test_asset2)

        # shots
        self.test_shot_status_list = \
            StatusList.query.filter_by(target_entity_type='Shot').first()

        # for project 1
        self.test_shot1 = Shot(
            code='shot1',
            project=self.test_project1,
            status_list=self.test_shot_status_list
        )
        DBSession.add(self.test_shot1)

        self.test_shot2 = Shot(
            code='shot2',
            project=self.test_project1,
            status_list=self.test_shot_status_list
        )
        DBSession.add(self.test_shot2)

        # for project 2
        self.test_shot3 = Shot(
            code='shot3',
            project=self.test_project2,
            status_list=self.test_shot_status_list
        )
        DBSession.add(self.test_shot3)

        self.test_shot4 = Shot(
            code='shot4',
            project=self.test_project2,
            status_list=self.test_shot_status_list
        )
        DBSession.add(self.test_shot4)

        # for project 3
        self.test_shot5 = Shot(
            code='shot5',
            project=self.test_project3,
            status_list=self.test_shot_status_list
        )
        DBSession.add(self.test_shot5)

        #########################################################3
        # tasks for projects
        self.test_task_statuses = \
            StatusList.query.filter_by(target_entity_type='Task').first()

        self.test_task1 = Task(
            name='Project Planing',
            project=self.test_project1,
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            schedule_timing=10,
            schedule_unit='d'
        )
        DBSession.add(self.test_task1)

        self.test_task2 = Task(
            name='Project Planing',
            project=self.test_project2,
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            schedule_timing=10,
            schedule_unit='d'
        )
        DBSession.add(self.test_task2)

        self.test_task3 = Task(
            name='Project Planing',
            project=self.test_project3,
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            status_list=self.test_task_statuses,
            schedule_timing=5,
            schedule_unit='d'
        )
        DBSession.add(self.test_task3)

        # for shots

        # Shot 1
        self.test_task4 = Task(
            name='Match Move',
            parent=self.test_shot1,
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task4)

        self.test_task5 = Task(
            name='FX',
            parent=self.test_shot1,
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            depends=[self.test_task4],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task5)

        self.test_task6 = Task(
            name='Lighting',
            parent=self.test_shot1,
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            depends=[self.test_task4, self.test_task5],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task6)

        self.test_task7 = Task(
            name='Comp',
            parent=self.test_shot1,
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            depends=[self.test_task6],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task7)

        # Shot 2
        self.test_task8 = Task(
            name='Match Move',
            parent=self.test_shot2,
            resources=[self.test_user3],
            alternative_resources=[self.test_user1, self.test_user2],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task8)

        self.test_task9 = Task(
            name='FX',
            parent=self.test_shot2,
            resources=[self.test_user3],
            alternative_resources=[self.test_user1, self.test_user2],
            depends=[self.test_task8],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task9)

        self.test_task10 = Task(
            name='Lighting',
            parent=self.test_shot2,
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            depends=[self.test_task8, self.test_task9],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task10)

        self.test_task11 = Task(
            name='Comp',
            parent=self.test_shot2,
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            depends=[self.test_task10],
            schedule_timing=4,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task11)

        # Shot 3
        self.test_task12 = Task(
            name='Match Move',
            parent=self.test_shot3,
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task12)

        self.test_task13 = Task(
            name='FX',
            parent=self.test_shot3,
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            depends=[self.test_task12],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task13)

        self.test_task14 = Task(
            name='Lighting',
            parent=self.test_shot3,
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            depends=[self.test_task12, self.test_task13],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task14)

        self.test_task15 = Task(
            name='Comp',
            parent=self.test_shot3,
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            depends=[self.test_task14],
            schedule_timing=4,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task15)

        # Shot 4
        self.test_task16 = Task(
            name='Match Move',
            parent=self.test_shot4,
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task16)

        self.test_task17 = Task(
            name='FX',
            parent=self.test_shot4,
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            depends=[self.test_task16],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task17)

        self.test_task18 = Task(
            name='Lighting',
            parent=self.test_shot4,
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            depends=[self.test_task16, self.test_task17],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task18)

        self.test_task19 = Task(
            name='Comp',
            parent=self.test_shot4,
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            depends=[self.test_task18],
            schedule_timing=4,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task19)

        # Shot 5
        self.test_task20 = Task(
            name='Match Move',
            parent=self.test_shot5,
            resources=[self.test_user3],
            alternative_resources=[self.test_user1, self.test_user2],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task20)

        self.test_task21 = Task(
            name='FX',
            parent=self.test_shot5,
            resources=[self.test_user3],
            alternative_resources=[self.test_user1, self.test_user2],
            depends=[self.test_task20],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task21)

        self.test_task22 = Task(
            name='Lighting',
            parent=self.test_shot5,
            resources=[self.test_user3],
            alternative_resources=[self.test_user1, self.test_user2],
            depends=[self.test_task20, self.test_task21],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task22)

        self.test_task23 = Task(
            name='Comp',
            parent=self.test_shot5,
            resources=[self.test_user3],
            alternative_resources=[self.test_user1, self.test_user2],
            depends=[self.test_task22],
            schedule_timing=4,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task23)

        ####################################################3
        # For Assets

        # Asset 1
        self.test_task24 = Task(
            name='Design',
            parent=self.test_asset1,
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task24)

        self.test_task25 = Task(
            name='Model',
            parent=self.test_asset1,
            depends=[self.test_task24],
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            schedule_timing=15,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task25)

        self.test_task26 = Task(
            name='LookDev',
            parent=self.test_asset1,
            depends=[self.test_task25],
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task26)

        self.test_task27 = Task(
            name='Rig',
            parent=self.test_asset1,
            depends=[self.test_task25],
            resources=[self.test_user1],
            alternative_resources=[self.test_user2, self.test_user3],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task27)

        # Asset 2
        self.test_task28 = Task(
            name='Design',
            parent=self.test_asset2,
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task28)

        self.test_task29 = Task(
            name='Model',
            parent=self.test_asset2,
            depends=[self.test_task28],
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            schedule_timing=15,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task29)

        self.test_task30 = Task(
            name='LookDev',
            parent=self.test_asset2,
            depends=[self.test_task29],
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task30)

        self.test_task31 = Task(
            name='Rig',
            parent=self.test_asset2,
            depends=[self.test_task29],
            resources=[self.test_user2],
            alternative_resources=[self.test_user1, self.test_user3],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task31)

        # TODO: Add Milestones

        self.kwargs = dict(
            name='Studio',
            daily_working_hours=8,
            timing_resolution=datetime.timedelta(hours=1)
        )

        self.test_studio = Studio(**self.kwargs)
        DBSession.add(self.test_studio)
        DBSession.commit()

    def tearDown(self):
        """clean up the test
        """
        DBSession.remove()

    def test_working_hours_argument_is_skipped(self):
        """testing if the default working hours will be used when the
        working_hours argument is skipped
        """
        self.kwargs['name'] = 'New Studio'
        try:
            self.kwargs.pop('working_hours')  # pop if there are any
        except KeyError:
            pass

        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.working_hours, WorkingHours())

    def test_working_hours_argument_is_None(self):
        """testing if the a WorkingHour instance with default settings will be
        used if the working_hours argument is skipped
        """
        self.kwargs['name'] = 'New Studio'
        self.kwargs['working_hours'] = None
        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.working_hours, WorkingHours())

    def test_working_hours_attribute_is_None(self):
        """testing if a WorkingHour instance will be created with the default
        values if the working_hours attribute is set to None
        """
        self.test_studio.working_horus = None
        self.assertEqual(self.test_studio.working_hours, WorkingHours())

    def test_tjp_id_attribute_returns_a_plausible_id(self):
        """testing if the tjp_id is returning something meaningful
        """
        self.test_studio.id = 432
        self.assertEqual(self.test_studio.tjp_id, 'Studio_432')

    def test_projects_attribute_is_read_only(self):
        """testing if the project attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio,
                          'projects', [self.test_project1])

    def test_projects_attribute_is_working_properly(self):
        """testing if the projects attribute is working properly
        """
        self.assertEqual(
            sorted(self.test_studio.projects, key=lambda x: x.name),
            sorted(
                [self.test_project1, self.test_project2, self.test_project3],
                key=lambda x: x.name
            )
        )

    def test_active_projects_attribute_is_read_only(self):
        """testing if the active_projects attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio,
                          'active_projects', [self.test_project1])

    def test_active_projects_attribute_is_working_properly(self):
        """testing if the active_projects attribute is working properly
        """
        self.assertEqual(
            sorted(self.test_studio.active_projects, key=lambda x: x.name),
            sorted([self.test_project1, self.test_project2],
                   key=lambda x: x.name)
        )

    def test_inactive_projects_attribute_is_read_only(self):
        """testing if the inactive_projects attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio,
                          'inactive_projects', [self.test_project1])

    def test_inactive_projects_attribute_is_working_properly(self):
        """testing if the inactive_projects attribute is working properly
        """
        self.assertEqual(
            sorted(self.test_studio.inactive_projects, key=lambda x: x.name),
            sorted([self.test_project3], key=lambda x: x.name)
        )

    def test_departments_attribute_is_read_only(self):
        """testing if the departments attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio,
                          'departments', [self.test_project1])

    def test_departments_attribute_is_working_properly(self):
        """testing if the departments attribute is working properly
        """
        # don't forget admins department
        db.init()
        admins_dep = Department.query.filter_by(name='admins').first()
        self.assertEqual(
            sorted(self.test_studio.departments, key=lambda x: x.name),
            sorted([self.test_department1, self.test_department2, admins_dep],
                   key=lambda x: x.name)
        )

    def test_users_attribute_is_read_only(self):
        """testing if the users attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio, 'users',
                          [self.test_project1])

    def test_users_attribute_is_working_properly(self):
        """testing if the users attribute is working properly
        """
        # don't forget the admin
        admin = User.query.filter_by(name='admin').first()

        self.assertEqual(
            sorted(self.test_studio.users, key=lambda x: x.name),
            sorted([admin, self.test_user1, self.test_user2, self.test_user3],
                   key=lambda x: x.name)
        )

    def test_to_tjp_attribute_is_read_only(self):
        """testing if the to_tjp attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio, 'to_tjp',
                          "some text")

    def test_now_argument_is_skipped(self):
        """testing if the now attribute will use the rounded
        datetime.datetime.now value when the now argument is skipped
        """
        try:
            self.kwargs.pop('now')
        except KeyError:
            pass

        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.now,
                         new_studio.round_time(datetime.datetime.now()))

    def test_now_argument_is_None(self):
        """testing if the now attribute will use the rounded
        datetime.datetime.now value when the now argument is None
        """
        self.kwargs['now'] = None
        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.now,
                         new_studio.round_time(datetime.datetime.now()))

    def test_now_attribute_is_None(self):
        """testing if the now attribute will be equal to the rounded value of
        datetime.datetime.now() if it is set to None
        """
        self.test_studio.now = None
        self.assertEqual(
            self.test_studio.now,
            self.test_studio.round_time(datetime.datetime.now())
        )

    def test_now_argument_is_not_a_datetime_instance(self):
        """testing if a TypeError will be raised when the now argument is not
        a datetime.datetime instance
        """
        self.kwargs['now'] = 'not a datetime instance'
        self.assertRaises(TypeError, Studio, **self.kwargs)

    def test_now_attribute_is_set_to_a_value_other_than_datetime_instance(self):
        """testing if a TypeError will be raised when the now attribute is set
        to a value other than a datetime.datetime instance
        """
        self.assertRaises(TypeError, setattr, self.test_studio, 'now',
                          'not a datetime instance')

    def test_now_argument_is_working_properly(self):
        """testing if the now argument value is passed to the now attribute
        properly
        """
        self.kwargs['now'] = datetime.datetime(2013, 4, 15, 21, 9)
        expected_now = datetime.datetime(2013, 4, 15, 21, 0)
        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.now, expected_now)

    def test_now_attribute_is_working_properly(self):
        """testing if the now attribute is working properly
        """
        self.test_studio.now = datetime.datetime(2013, 4, 15, 21, 11)
        expected_now = datetime.datetime(2013, 4, 15, 21, 0)
        self.assertEqual(self.test_studio.now, expected_now)

    def test_to_tjp_attribute_is_working_properly(self):
        """testing if the to_tjp attribute is working properly
        """
        self.test_studio.start = datetime.datetime(2013, 4, 15, 17, 40)
        self.test_studio.end = datetime.datetime(2013, 6, 30, 17, 40)
        now = self.test_studio.round_time(datetime.datetime.now())
        self.test_studio.working_hours[0] = [[540, 1080]]
        self.test_studio.working_hours[1] = [[540, 1080]]
        self.test_studio.working_hours[2] = [[540, 1080]]
        self.test_studio.working_hours[3] = [[540, 1080]]
        self.test_studio.working_hours[4] = [[540, 1080]]
        self.test_studio.working_hours[5] = [[540, 720]]
        self.test_studio.working_hours[6] = []

        from jinja2 import Template

        expected_tjp_template = Template("""
project Studio_{{ studio.id }} "Studio" 2013-04-15 - 2013-06-30 {
    timingresolution 60min
    now {{ studio.now.strftime('%Y-%m-%d-%H:%M') }}
    dailyworkinghours 8
    weekstartsmonday
    workinghours mon 09:00 - 18:00
    workinghours tue 09:00 - 18:00
    workinghours wed 09:00 - 18:00
    workinghours thu 09:00 - 18:00
    workinghours fri 09:00 - 18:00
    workinghours sat 09:00 - 12:00
    workinghours sun off
    timeformat "%Y-%m-%d"
    scenario plan "Plan"
    trackingscenario plan
}
""")

        expected_tjp = expected_tjp_template.render({
            'studio': self.test_studio
        })
        # print('-----------------------------------')
        # print(expected_tjp)
        # print('-----------------------------------')
        # print(self.test_studio.to_tjp)
        # print('-----------------------------------')
        self.assertEqual(self.test_studio.to_tjp, expected_tjp)

    def test_scheduler_attribute_can_be_set_to_None(self):
        """testing if the scheduler attribute can be set to None
        """
        self.test_studio.scheduler = None

    def test_scheduler_attribute_accepts_Scheduler_instances_only(self):
        """testing if a TypeError will be raised when the scheduler attribute
        is set to a value which is not a scheduler instance
        """
        self.assertRaises(TypeError, setattr, self.test_studio, 'scheduler',
                          'not a Scheduler instance')

    def test_scheduler_attribute_is_working_properly(self):
        """testing if the scheduler attribute is working properly
        """
        tj_s = TaskJugglerScheduler()
        self.test_studio.scheduler = tj_s
        self.assertEqual(self.test_studio.scheduler, tj_s)

    def test_schedule_will_not_work_without_a_scheduler(self):
        """testing if a RuntimeError will be raised when the scheduler
        attribute is not set to a Scheduler instance and schedule is called
        """
        self.test_studio.scheduler = None
        self.assertRaises(RuntimeError, self.test_studio.schedule)

    def test_schedule_will_schedule_the_tasks_with_the_given_scheduler(self):
        """testing if the schedule method will schedule the tasks with the
        given scheduler
        """
        tj_scheduler = TaskJugglerScheduler(compute_resources=True)
        self.test_studio.now = datetime.datetime(2013, 4, 15, 22, 56)
        self.test_studio.start = datetime.datetime(2013, 4, 15, 22, 56)
        self.test_studio.end = datetime.datetime(2013, 7, 30, 0, 0)

        # just to be sure that it is not creating any issue on schedule
        self.test_task25.task_depends_to[0].dependency_target = 'onstart'
        self.test_task25.resources = [self.test_user2]

        self.test_studio.scheduler = tj_scheduler
        self.test_studio.schedule()
        db.DBSession.commit()

        # now check the timings of the tasks are all adjusted

        # Projects
        # print "%s:self.test_project1" % self.test_project1.id
        # print "%s:self.test_project2" % self.test_project2.id
        # print "%s:self.test_project3" % self.test_project3.id
        # 
        # print "%s:self.test_asset1" % self.test_asset1.id
        # print "%s:self.test_asset2" % self.test_asset2.id
        # 
        # print "%s:self.test_shot1.id" % self.test_shot1.id
        # print "%s:self.test_shot2.id" % self.test_shot2.id
        # print "%s:self.test_shot3.id" % self.test_shot3.id
        # print "%s:self.test_shot4.id" % self.test_shot4.id
        # print "%s:self.test_shot5.id" % self.test_shot5.id
        # 
        # print "%s:self.test_task1.id" % self.test_task1.id
        # print "%s:self.test_task2.id" % self.test_task2.id
        # print "%s:self.test_task3.id" % self.test_task3.id
        # print "%s:self.test_task4.id" % self.test_task4.id
        # print "%s:self.test_task5.id" % self.test_task5.id
        # print "%s:self.test_task6.id" % self.test_task6.id
        # print "%s:self.test_task7.id" % self.test_task7.id
        # print "%s:self.test_task8.id" % self.test_task8.id
        # print "%s:self.test_task9.id" % self.test_task9.id
        # 
        # print "%s:self.test_task10.id" % self.test_task10.id
        # print "%s:self.test_task11.id" % self.test_task11.id
        # print "%s:self.test_task12.id" % self.test_task12.id
        # print "%s:self.test_task13.id" % self.test_task13.id
        # print "%s:self.test_task14.id" % self.test_task14.id
        # print "%s:self.test_task15.id" % self.test_task15.id
        # print "%s:self.test_task16.id" % self.test_task16.id
        # print "%s:self.test_task17.id" % self.test_task17.id
        # print "%s:self.test_task18.id" % self.test_task18.id
        # print "%s:self.test_task19.id" % self.test_task19.id
        # 
        # print "%s:self.test_task20.id" % self.test_task20.id
        # print "%s:self.test_task21.id" % self.test_task21.id
        # print "%s:self.test_task22.id" % self.test_task22.id
        # print "%s:self.test_task23.id" % self.test_task23.id
        # print "%s:self.test_task24.id" % self.test_task24.id
        # print "%s:self.test_task25.id" % self.test_task25.id
        # print "%s:self.test_task26.id" % self.test_task26.id
        # print "%s:self.test_task27.id" % self.test_task27.id
        # print "%s:self.test_task28.id" % self.test_task28.id
        # print "%s:self.test_task29.id" % self.test_task29.id
        # print "%s:self.test_task30.id" % self.test_task30.id
        # print "%s:self.test_task31.id" % self.test_task31.id

        # self.test_project1
        self.assertEqual(
            datetime.datetime(2013, 4, 16, 9, 0),
            self.test_project1.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 19, 11, 0),
            self.test_project1.computed_end
        )

        # self.test_asset1
        self.assertEqual(
            datetime.datetime(2013, 4, 16, 9, 0),
            self.test_asset1.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 17, 10, 0),
            self.test_asset1.computed_end
        )

        self.assertEqual(
            self.test_asset1.computed_resources,
            []
        )

        # self.test_task24
        self.assertEqual(
            datetime.datetime(2013, 4, 16, 9, 0),
            self.test_task24.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 4, 26, 17, 0),
            self.test_task24.computed_end
        )

        self.assertEqual(
            self.test_task24.computed_resources,
            [self.test_user1]
        )

        # self.test_task25
        self.assertEqual(
            datetime.datetime(2013, 4, 16, 9, 0),
            self.test_task25.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 3, 12, 0),
            self.test_task25.computed_end
        )

        self.assertEqual(
            self.test_task25.computed_resources,
            [self.test_user3]
        )

        # self.test_task26
        self.assertEqual(
            datetime.datetime(2013, 5, 3, 12, 0),
            self.test_task26.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 16, 11, 0),
            self.test_task26.computed_end
        )

        self.assertEqual(
            self.test_task26.computed_resources,
            [self.test_user3]
        )

        # self.test_task27
        self.assertEqual(
            datetime.datetime(2013, 5, 6, 11, 0),
            self.test_task27.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 17, 10, 0),
            self.test_task27.computed_end
        )

        self.assertEqual(
            self.test_task27.computed_resources,
            [self.test_user2]
        )

        # self.test_shot2
        self.assertEqual(
            datetime.datetime(2013, 4, 26, 17, 0),
            self.test_shot2.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 14, 14, 0),
            self.test_shot2.computed_end
        )

        self.assertEqual(
            self.test_shot2.computed_resources,
            []
        )

        # self.test_task8
        self.assertEqual(
            datetime.datetime(2013, 4, 26, 17, 0),
            self.test_task8.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 4, 30, 15, 0),
            self.test_task8.computed_end
        )

        self.assertEqual(
            self.test_task8.computed_resources,
            [self.test_user2]
        )

        # self.test_task9
        self.assertEqual(
            datetime.datetime(2013, 5, 30, 9, 0),
            self.test_task9.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 31, 16, 0),
            self.test_task9.computed_end
        )

        self.assertEqual(
            self.test_task9.computed_resources,
            [self.test_user2]
        )

        # self.test_task10
        self.assertEqual(
            datetime.datetime(2013, 6, 4, 14, 0),
            self.test_task10.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 7, 11, 0),
            self.test_task10.computed_end
        )

        self.assertEqual(
            self.test_task10.computed_resources,
            [self.test_user2]
        )

        # self.test_task11
        self.assertEqual(
            datetime.datetime(2013, 6, 11, 9, 0),
            self.test_task11.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 14, 14, 0),
            self.test_task11.computed_end
        )

        self.assertEqual(
            self.test_task11.computed_resources,
            [self.test_user3]
        )

        # self.test_shot1
        self.assertEqual(
            datetime.datetime(2013, 5, 16, 11, 0),
            self.test_shot1.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 19, 11, 0),
            self.test_shot1.computed_end
        )

        self.assertEqual(
            self.test_shot1.computed_resources,
            []
        )

        # self.test_task4
        self.assertEqual(
            datetime.datetime(2013, 5, 16, 11, 0),
            self.test_task4.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 17, 18, 0),
            self.test_task4.computed_end
        )

        self.assertEqual(
            self.test_task4.computed_resources,
            [self.test_user1]
        )

        # self.test_task5
        self.assertEqual(
            datetime.datetime(2013, 6, 3, 15, 0),
            self.test_task5.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 5, 13, 0),
            self.test_task5.computed_end
        )

        self.assertEqual(
            self.test_task5.computed_resources,
            [self.test_user1]
        )

        # self.test_task6
        self.assertEqual(
            datetime.datetime(2013, 6, 10, 10, 0),
            self.test_task6.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 12, 16, 0),
            self.test_task6.computed_end
        )

        self.assertEqual(
            self.test_task6.computed_resources,
            [self.test_user1]
        )

        # self.test_task7
        self.assertEqual(
            datetime.datetime(2013, 6, 14, 14, 0),
            self.test_task7.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 19, 11, 0),
            self.test_task7.computed_end
        )

        self.assertEqual(
            self.test_task7.computed_resources,
            [self.test_user3]
        )

        # self.test_task1
        self.assertEqual(
            datetime.datetime(2013, 5, 16, 11, 0),
            self.test_task1.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 29, 10, 0),
            self.test_task1.computed_end
        )

        self.assertEqual(
            self.test_task1.computed_resources,
            [self.test_user3]
        )

        # self.test_project2
        # self.assertEqual(
        #     datetime.datetime(2013, 4, 16, 9, 0),
        #     self.test_project2.computed_start
        # )
        # self.assertEqual(
        #     datetime.datetime(2013, 6, 18, 12, 0),
        #     self.test_project2.computed_end
        # )

        # self.assertEqual(
        #     self.test_project2.computed_resources,
        #     []
        # )

        # self.test_asset2
        self.assertEqual(
            datetime.datetime(2013, 4, 16, 9, 0),
            self.test_asset2.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 30, 17, 0),
            self.test_asset2.computed_end
        )

        self.assertEqual(
            self.test_asset2.computed_resources,
            []
        )

        # self.test_task28
        self.assertEqual(
            datetime.datetime(2013, 4, 16, 9, 0),
            self.test_task28.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 4, 26, 17, 0),
            self.test_task28.computed_end
        )

        self.assertEqual(
            self.test_task28.computed_resources,
            [self.test_user2]
        )

        # self.test_task29
        self.assertEqual(
            datetime.datetime(2013, 4, 26, 17, 0),
            self.test_task29.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 16, 11, 0),
            self.test_task29.computed_end
        )

        self.assertEqual(
            self.test_task29.computed_resources,
            [self.test_user1]
        )

        # self.test_task30
        self.assertEqual(
            datetime.datetime(2013, 5, 17, 10, 0),
            self.test_task30.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 29, 18, 0),
            self.test_task30.computed_end
        )

        self.assertEqual(
            self.test_task30.computed_resources,
            [self.test_user2]
        )

        # self.test_task31
        self.assertEqual(
            datetime.datetime(2013, 5, 20, 9, 0),
            self.test_task31.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 30, 17, 0),
            self.test_task31.computed_end
        )

        self.assertEqual(
            self.test_task31.computed_resources,
            [self.test_user1]
        )

        # self.test_shot3
        self.assertEqual(
            datetime.datetime(2013, 4, 30, 15, 0),
            self.test_shot3.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 17, 13, 0),
            self.test_shot3.computed_end
        )

        self.assertEqual(
            self.test_shot3.computed_resources,
            []
        )

        # self.test_task12
        self.assertEqual(
            datetime.datetime(2013, 4, 30, 15, 0),
            self.test_task12.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 2, 13, 0),
            self.test_task12.computed_end
        )

        self.assertEqual(
            self.test_task12.computed_resources,
            [self.test_user2]
        )

        # self.test_task13
        self.assertEqual(
            datetime.datetime(2013, 5, 30, 17, 0),
            self.test_task13.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 3, 15, 0),
            self.test_task13.computed_end
        )

        self.assertEqual(
            self.test_task13.computed_resources,
            [self.test_user1]
        )

        # self.test_task14
        self.assertEqual(
            datetime.datetime(2013, 6, 5, 13, 0),
            self.test_task14.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 10, 10, 0),
            self.test_task14.computed_end
        )

        self.assertEqual(
            self.test_task14.computed_resources,
            [self.test_user1]
        )

        # self.test_task15
        self.assertEqual(
            datetime.datetime(2013, 6, 11, 17, 0),
            self.test_task15.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 17, 13, 0),
            self.test_task15.computed_end
        )

        self.assertEqual(
            self.test_task15.computed_resources,
            [self.test_user2]
        )

        # self.test_shot4
        self.assertEqual(
            datetime.datetime(2013, 5, 2, 13, 0),
            self.test_shot4.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 18, 12, 0),
            self.test_shot4.computed_end
        )

        self.assertEqual(
            self.test_shot4.computed_resources,
            []
        )

        # self.test_task16
        self.assertEqual(
            datetime.datetime(2013, 5, 2, 13, 0),
            self.test_task16.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 5, 6, 11, 0),
            self.test_task16.computed_end
        )

        self.assertEqual(
            self.test_task16.computed_resources,
            [self.test_user2]
        )

        # self.test_task17
        self.assertEqual(
            datetime.datetime(2013, 5, 31, 16, 0),
            self.test_task17.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 4, 14, 0),
            self.test_task17.computed_end
        )

        self.assertEqual(
            self.test_task17.computed_resources,
            [self.test_user2]
        )

        # self.test_task18
        self.assertEqual(
            datetime.datetime(2013, 6, 7, 11, 0),
            self.test_task18.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 11, 17, 0),
            self.test_task18.computed_end
        )

        self.assertEqual(
            self.test_task18.computed_resources,
            [self.test_user2]
        )

        # self.test_task19
        self.assertEqual(
            datetime.datetime(2013, 6, 12, 16, 0),
            self.test_task19.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 18, 12, 0),
            self.test_task19.computed_end
        )

        self.assertEqual(
            self.test_task19.computed_resources,
            [self.test_user1]
        )

        # self.test_task2
        self.assertEqual(
            datetime.datetime(2013, 5, 29, 10, 0),
            self.test_task2.computed_start
        )
        self.assertEqual(
            datetime.datetime(2013, 6, 10, 18, 0),
            self.test_task2.computed_end
        )

        self.assertEqual(
            self.test_task2.computed_resources,
            [self.test_user3]
        )

    # def test_schedule_will_raise_a_RuntimeError_if_is_scheduling_is_True(self):
    #     """testing if a RuntimeError will be raised when the schedule method
    #     is called and the is_scheduling attribute is True
    #     """
    #     tj_scheduler = TaskJugglerScheduler()
    #     self.test_studio.now = datetime.datetime(2013, 4, 15, 22, 56)
    #     self.test_studio.start = datetime.datetime(2013, 4, 15, 22, 56)
    #     self.test_studio.end = datetime.datetime(2013, 7, 30, 0, 0)
    # 
    #     self.test_studio.scheduler = tj_scheduler
    #     self.test_studio.is_scheduling = True
    #     self.test_studio.is_scheduling_by = self.test_user1
    #     self.assertRaises(RuntimeError, self.test_studio.schedule)

    def test_is_scheduling_will_be_False_after_scheduling_is_done(self):
        """testing if the is_scheduling attribute will be back to False when
        the scheduling is finished
        """
        # use a dummy scheduler
        self.test_studio.now = datetime.datetime(2013, 4, 15, 22, 56)
        self.test_studio.start = datetime.datetime(2013, 4, 15, 22, 56)
        self.test_studio.end = datetime.datetime(2013, 7, 30, 0, 0)

        def callback():
            self.assertTrue(self.test_studio.is_scheduling)

        dummy_scheduler = DummyScheduler(callback=callback)

        self.test_studio.scheduler = dummy_scheduler
        self.assertFalse(self.test_studio.is_scheduling)

        # with v0.2.6.9 it is now the users duty to set is_scheduling to True
        self.test_studio.is_scheduling = True

        self.test_studio.schedule()
        self.assertFalse(self.test_studio.is_scheduling)

    def test_schedule_will_store_schedule_info_in_database(self):
        """testing if the schedule method will store the schedule info in
        database
        """
        tj_scheduler = TaskJugglerScheduler()
        self.test_studio.now = datetime.datetime(2013, 4, 15, 22, 56)
        self.test_studio.start = datetime.datetime(2013, 4, 15, 22, 56)
        self.test_studio.end = datetime.datetime(2013, 7, 30, 0, 0)

        self.test_studio.scheduler = tj_scheduler
        self.test_studio.schedule(scheduled_by=self.test_user1)

        self.assertEqual(self.test_studio.last_scheduled_by, self.test_user1)

        last_schedule_message = self.test_studio.last_schedule_message
        last_scheduled_at = self.test_studio.last_scheduled_at
        last_scheduled_by = self.test_studio.last_scheduled_by

        self.assertIsNotNone(last_schedule_message)
        self.assertIsNotNone(last_scheduled_at)
        self.assertIsNotNone(last_scheduled_by)

        db.DBSession.add(self.test_studio)
        db.DBSession.commit()

        # delete the studio instance and retrieve it back and check if it has
        # the info
        del self.test_studio

        studio = Studio.query.first()

        self.assertFalse(studio.is_scheduling)
        self.assertTrue(
            datetime.datetime.now() - studio.scheduling_started_at <
            datetime.timedelta(minutes=1)
        )
        self.assertEqual(last_schedule_message, studio.last_schedule_message)
        self.assertEqual(last_scheduled_at, studio.last_scheduled_at)
        self.assertEqual(last_scheduled_by, studio.last_scheduled_by)

        self.assertEqual(studio.last_scheduled_by_id, self.test_user1.id)
        self.assertEqual(studio.last_scheduled_by, self.test_user1)

    def test_vacation_attribute_is_read_only(self):
        """testing if the vacation attribute is a read-only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio,
                          'vacations', 'some random value')

    def test_vacation_attribute_returns_studio_vacation_instances(self):
        """Testing if the vacation attribute is returning the Vacation
        instances with no user set.
        """
        from stalker.models.studio import Vacation

        vacation1 = Vacation(
            start=datetime.datetime(2013, 8, 2),
            end=datetime.datetime(2013, 8, 10)
        )
        vacation2 = Vacation(
            start=datetime.datetime(2013, 8, 11),
            end=datetime.datetime(2013, 8, 20)
        )
        vacation3 = Vacation(
            user=self.test_user1,
            start=datetime.datetime(2013, 8, 11),
            end=datetime.datetime(2013, 8, 20)
        )
        DBSession.add_all([vacation1, vacation2, vacation3])
        DBSession.commit()

        self.assertEqual(
            sorted(self.test_studio.vacations, key=lambda x: x.name),
            sorted([vacation1, vacation2], key=lambda x: x.name)
        )

    def test_timing_resolution_argument_skipped(self):
        """testing if the timing_resolution attribute will be set to the
        default value from the defaults.timing_resolution if timing_resolution
        argument is skipped
        """
        try:
            self.kwargs.pop('timing_resolution')
        except KeyError:
            pass

        studio = Studio(**self.kwargs)
        self.assertEqual(
            studio.timing_resolution,
            defaults.timing_resolution
        )

    # def test_timing_resolution_argument_skipped_Studio_is_present(self):
    #     """testing if the timing_resolution attribute will be set to the Studio
    #     timing resolution if timing_resolution argument is skipped and there is
    #     a Studio instance
    #     """
    #     try:
    #         self.kwargs.pop('timing_resolution')
    #     except KeyError:
    #         pass
    # 
    #     from stalker import Studio
    #     db.setup({'sqlalchemy.url': 'sqlite:///:memory:'})
    #     studio = Studio(
    #         name='Test Studio',
    #         timing_resolution=datetime.timedelta(minutes=15)
    #     )
    #     DBSession.add(studio)
    #     DBSession.commit()
    # 
    #     new_foo_obj = DateRangeMixFooMixedInClass(**self.kwargs)
    #     self.assertEqual(
    #         new_foo_obj.timing_resolution,
    #         studio.timing_resolution
    #     )

    def test_timing_resolution_argument_is_None(self):
        """testing if the timing_resolution attribute will be set to the
        default value from the default.timing_resolution if timing_resolution
        argument is None
        """
        self.kwargs['timing_resolution'] = None
        studio = Studio(**self.kwargs)
        self.assertEqual(
            studio.timing_resolution,
            defaults.timing_resolution
        )

    # def test_timing_resolution_argument_is_None_Studio_is_present(self):
    #     """testing if the timing_resolution attribute will be set to the
    #     default value from the Studio.timing_resolution if timing_resolution
    #     argument is None and there is a Studio
    #     """
    #     self.kwargs['timing_resolution'] = None
    # 
    #     from stalker import Studio
    #     db.setup({'sqlalchemy.url': 'sqlite:///:memory:'})
    #     studio = Studio(
    #         name='Test Studio',
    #         timing_resolution=datetime.timedelta(minutes=15)
    #     )
    #     DBSession.add(studio)
    #     DBSession.commit()
    # 
    #     new_foo_obj = DateRangeMixFooMixedInClass(**self.kwargs)
    #     self.assertEqual(
    #         new_foo_obj.timing_resolution,
    #         studio.timing_resolution
    #     )

    def test_timing_resolution_attribute_is_set_to_None(self):
        """testing if the timing_resolution attribute will be set to the
        default value from the defaults.timing_resolution if it is set to None
        """
        self.kwargs['timing_resolution'] = datetime.timedelta(minutes=5)
        studio = Studio(**self.kwargs)
        # check start conditions
        self.assertEqual(
            studio.timing_resolution,
            self.kwargs['timing_resolution']
        )
        studio.timing_resolution = None
        self.assertEqual(
            studio.timing_resolution,
            defaults.timing_resolution
        )

    # def test_timing_resolution_attribute_is_set_to_None_Studio_is_present(self):
    #     """testing if the timing_resolution attribute will be set to the
    #     default value from the Studio.timing_resolution if it is set to None
    #     and there is a Studio instance
    #     """
    #     from stalker import Studio
    #     db.setup({'sqlalchemy.url': 'sqlite:///:memory:'})
    #     studio = Studio(
    #         name='Test Studio',
    #         resolution=datetime.timedelta(minutes=15)
    #     )
    #     session.add(studio)
    #     session.commit()
    # 
    #     self.kwargs['timing_resolution'] = datetime.timedelta(minutes=5)
    #     new_foo_obj = DateRangeMixFooMixedInClass(**self.kwargs)
    #     # check start conditions
    #     self.assertEqual(new_foo_obj.timing_resolution,
    #                      self.kwargs['timing_resolution'])
    #     new_foo_obj.timing_resolution = None
    #     self.assertEqual(
    #         studio.timing_resolution,
    #         datetime.timedelta(minutes=15)
    #     )
    #     self.assertEqual(
    #         new_foo_obj.timing_resolution,
    #         studio.timing_resolution
    #     )

    def test_timing_resolution_argument_is_not_a_timedelta_instance(self):
        """testing if a TypeError will be raised when the timing_resolution
        argument is not a datetime.timedelta instance
        """
        self.kwargs['timing_resolution'] = 'not a timedelta instance'
        self.assertRaises(TypeError, Studio, **self.kwargs)

    def test_timing_resolution_attribute_is_not_a_timedelta_instance(self):
        """testing if a TypeError will be raised when the timing_resolution
        attribute is not a datetime.timedelta instance
        """
        new_foo_obj = Studio(**self.kwargs)
        self.assertRaises(TypeError, setattr, new_foo_obj, 'timing_resolution',
                          'not a timedelta instance')

    def test_timing_resolution_argument_is_working_properly(self):
        """testing if the timing_resolution argument value is passed to
        timing_resolution attribute correctly
        """
        self.kwargs['timing_resolution'] = datetime.timedelta(minutes=5)
        studio = Studio(**self.kwargs)
        self.assertEqual(
            studio.timing_resolution,
            self.kwargs['timing_resolution']
        )

    def test_timing_resolution_attribute_is_working_properly(self):
        """testing if the timing_resolution attribute is working properly
        """
        studio = Studio(**self.kwargs)
        res = studio
        new_res = datetime.timedelta(hours=1, minutes=30)
        self.assertNotEqual(res, new_res)
        studio.timing_resolution = new_res
        self.assertEqual(studio.timing_resolution, new_res)


@unittest.skip
def csv_to_test_converter():
    """convert tjp output csv to test case
    """
    #!/usr/bin/python

    import re
    import jinja2

    ids = """28:self.test_user1
    29:self.test_user2
    30:self.test_user3
    38:self.test_project1
    39:self.test_project2
    40:self.test_project3
    41:self.test_asset1
    42:self.test_asset2
    43:self.test_shot1
    44:self.test_shot2
    45:self.test_shot3
    46:self.test_shot4
    47:self.test_shot5
    48:self.test_task1
    49:self.test_task2
    50:self.test_task3
    51:self.test_task4
    52:self.test_task5
    53:self.test_task6
    54:self.test_task7
    55:self.test_task8
    56:self.test_task9
    57:self.test_task10
    58:self.test_task11
    59:self.test_task12
    60:self.test_task13
    61:self.test_task14
    62:self.test_task15
    63:self.test_task16
    64:self.test_task17
    65:self.test_task18
    66:self.test_task19
    67:self.test_task20
    68:self.test_task21
    69:self.test_task22
    70:self.test_task23
    71:self.test_task24
    72:self.test_task25
    73:self.test_task26
    74:self.test_task27
    75:self.test_task28
    76:self.test_task29
    77:self.test_task30
    78:self.test_task31"""

    ids_to_name = {}
    for line in ids.split('\n'):
        id_, name = line.split(':')
        ids_to_name[id_] = name

    raw_data = '''"Project_38";"2013-04-16-09:00";"2013-06-19-11:00";""
    "Project_38.Asset_41";"2013-04-16-09:00";"2013-05-17-10:00";""
    "Project_38.Asset_41.Task_71";"2013-04-16-09:00";"2013-04-26-17:00";"User 1 (User_28)"
    "Project_38.Asset_41.Task_72";"2013-04-16-09:00";"2013-05-03-12:00";"User 3 (User_30)"
    "Project_38.Asset_41.Task_73";"2013-05-03-12:00";"2013-05-16-11:00";"User 3 (User_30)"
    "Project_38.Asset_41.Task_74";"2013-05-06-11:00";"2013-05-17-10:00";"User 2 (User_29)"
    "Project_38.Shot_44";"2013-04-26-17:00";"2013-06-14-14:00";""
    "Project_38.Shot_44.Task_55";"2013-04-26-17:00";"2013-04-30-15:00";"User 2 (User_29)"
    "Project_38.Shot_44.Task_56";"2013-05-30-09:00";"2013-05-31-16:00";"User 2 (User_29)"
    "Project_38.Shot_44.Task_57";"2013-06-04-14:00";"2013-06-07-11:00";"User 2 (User_29)"
    "Project_38.Shot_44.Task_58";"2013-06-11-09:00";"2013-06-14-14:00";"User 3 (User_30)"
    "Project_38.Shot_43";"2013-05-16-11:00";"2013-06-19-11:00";""
    "Project_38.Shot_43.Task_51";"2013-05-16-11:00";"2013-05-17-18:00";"User 1 (User_28)"
    "Project_38.Shot_43.Task_52";"2013-06-03-15:00";"2013-06-05-13:00";"User 1 (User_28)"
    "Project_38.Shot_43.Task_53";"2013-06-10-10:00";"2013-06-12-16:00";"User 1 (User_28)"
    "Project_38.Shot_43.Task_54";"2013-06-14-14:00";"2013-06-19-11:00";"User 3 (User_30)"
    "Project_38.Task_48";"2013-05-16-11:00";"2013-05-29-10:00";"User 3 (User_30)"
    "Project_39";"2013-04-16-09:00";"2013-06-18-12:00";""
    "Project_39.Asset_42";"2013-04-16-09:00";"2013-05-30-17:00";""
    "Project_39.Asset_42.Task_75";"2013-04-16-09:00";"2013-04-26-17:00";"User 2 (User_29)"
    "Project_39.Asset_42.Task_76";"2013-04-26-17:00";"2013-05-16-11:00";"User 1 (User_28)"
    "Project_39.Asset_42.Task_77";"2013-05-17-10:00";"2013-05-29-18:00";"User 2 (User_29)"
    "Project_39.Asset_42.Task_78";"2013-05-20-09:00";"2013-05-30-17:00";"User 1 (User_28)"
    "Project_39.Shot_45";"2013-04-30-15:00";"2013-06-17-13:00";""
    "Project_39.Shot_45.Task_59";"2013-04-30-15:00";"2013-05-02-13:00";"User 2 (User_29)"
    "Project_39.Shot_45.Task_60";"2013-05-30-17:00";"2013-06-03-15:00";"User 1 (User_28)"
    "Project_39.Shot_45.Task_61";"2013-06-05-13:00";"2013-06-10-10:00";"User 1 (User_28)"
    "Project_39.Shot_45.Task_62";"2013-06-11-17:00";"2013-06-17-13:00";"User 2 (User_29)"
    "Project_39.Shot_46";"2013-05-02-13:00";"2013-06-18-12:00";""
    "Project_39.Shot_46.Task_63";"2013-05-02-13:00";"2013-05-06-11:00";"User 2 (User_29)"
    "Project_39.Shot_46.Task_64";"2013-05-31-16:00";"2013-06-04-14:00";"User 2 (User_29)"
    "Project_39.Shot_46.Task_65";"2013-06-07-11:00";"2013-06-11-17:00";"User 2 (User_29)"
    "Project_39.Shot_46.Task_66";"2013-06-12-16:00";"2013-06-18-12:00";"User 1 (User_28)"
    "Project_39.Task_49";"2013-05-29-10:00";"2013-06-10-18:00";"User 3 (User_30)"'''

    lines = raw_data.split('\n')

    template_string = """        # {{entity_name}}
            self.assertEqual(
                datetime.datetime({{start_date}}),
                {{entity_name}}.computed_start
            )
            self.assertEqual(
                datetime.datetime({{end_date}}),
                {{entity_name}}.computed_end
            )
            {% if resources %}
            self.assertEqual(
                sorted({{entity_name}}.computed_resources, key=lambda x: x.name),
                soretd([{% for r in resources -%}
                {{r}}{% if loop.index != 1%},{% endif %}
                {%- endfor %}], key=lambda x: x.name)
            )
            {% endif %}
    """

    rendered_data = []
    template = jinja2.Template(template_string)
    for i, line in enumerate(lines):
        line = re.subn('[\s]+', ' ', line)[0]
        line = line.replace('"', '')

        data = line.split(';')

        entity_id = data[0].split('_')[-1]
        entity_name = ids_to_name[entity_id]

        start_date = ', '.join(
            data[1]
            .replace(':', '-')
            .replace('-0', '-')
            .replace('00', '0')
            .split('-')
        )

        end_date = ', '.join(
            data[2]
            .replace(':', '-')
            .replace('-0', '-')
            .replace('00', '0')
            .split('-')
        )

        resource_id = data[3].split('_')[-1][:-1]
        resource_name = ''
        if resource_id:
            resource_name = ids_to_name[resource_id]

        resources = [resource_name]

        #print entity_name, start_date, end_date, resources

        rendered_template = template.render(**{
            'entity_name': entity_name,
            'start_date': start_date,
            'end_date': end_date,
            'resources': resources
        })
        rendered_data.append(rendered_template)

    for r in rendered_data:
        print(r)



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest

from sqlalchemy import Column, Integer, ForeignKey
from stalker.models.entity import Entity
from stalker.models.link import Link
from stalker.models.mixins import ReferenceMixin
from stalker.db.session import DBSession
from stalker.models.type import Type
from stalker.models.entity import SimpleEntity


class RefMixFooClass(SimpleEntity, ReferenceMixin):
    __tablename__ = "RefMixFooClasses"
    __mapper_args__ = {"polymorphic_identity": "RefMixFooClass"}
    refMixFooClass_id = Column("id", Integer, ForeignKey("SimpleEntities.id"),
                               primary_key=True)

    def __init__(self, **kwargs):
        super(RefMixFooClass, self).__init__(**kwargs)


class ReferenceMixinTester(unittest.TestCase):
    """tests the ReferenceMixin
    """

    def setUp(self):
        """setup the test
        """
        # link type
        self.test_link_type = Type(
            name='Test Link Type',
            code='testlink',
            target_entity_type=Link,
        )

        # create a couple of Link objects
        self.test_link1 = Link(
            name="Test Link 1",
            type=self.test_link_type,
            full_path="test_path",
            filename="test_filename",
        )

        self.test_link2 = Link(
            name="Test Link 2",
            type=self.test_link_type,
            full_path="test_path",
            filename="test_filename",
        )

        self.test_link3 = Link(
            name="Test Link 3",
            type=self.test_link_type,
            full_path="test_path",
            filename="test_filename",
        )

        self.test_link4 = Link(
            name="Test Link 4",
            type=self.test_link_type,
            full_path="test_path",
            filename="test_filename",
        )

        self.test_entity1 = Entity(
            name="Test Entity 1",
        )

        self.test_entity2 = Entity(
            name="Test Entity 2",
        )

        self.test_links = [
            self.test_link1,
            self.test_link2,
            self.test_link3,
            self.test_link4,
        ]

        self.test_foo_obj = RefMixFooClass(name="Ref Mixin Test")

    def tearDown(self):
        """clean up the test
        """
        DBSession.remove()

    def test_references_attribute_accepting_empty_list(self):
        """testing if references attribute accepting empty lists
        """
        self.test_foo_obj.references = []

    def test_references_attribute_only_accepts_listlike_objects(self):
        """testing if references attribute accepts only list-like objects,
        (objects with __setitem__, __getitem__ methods
        """
        test_values = [1, 1.2, "a string"]

        for test_value in test_values:
            self.assertRaises(
                TypeError,
                setattr,
                self.test_foo_obj,
                "references",
                test_value
            )

    def test_references_attribute_accepting_only_lists_of_Link_instances(self):
        """testing if references attribute accepting only lists with Link
        instances and derivatives
        """
        test_value = [1, 2.2, ["a reference as list"], "some references"]

        self.assertRaises(
            TypeError,
            setattr,
            self.test_foo_obj,
            "references",
            test_value
        )

        # and test if it is accepting a proper list
        test_value = [self.test_link1, self.test_link2, self.test_link3]
        self.test_foo_obj.references = test_value
        self.assertEqual(
            sorted(test_value, key=lambda x: x.name),
            sorted(self.test_foo_obj.references, key=lambda x: x.name)
        )

    def test_references_attribute_elements_accepts_Links_only(self):
        """testing if a TypeError will be raised when trying to assign
        something other than an instance of Link or its derived classes to
        the references list
        """
        self.assertRaises(
            TypeError,
            setattr,
            self.test_foo_obj, 'references',
            [self.test_entity1, self.test_entity2]
        )

    def test_references_attribute_working_properly(self):
        """testing if references attribute working properly
        """
        self.test_foo_obj.references = self.test_links
        self.assertEqual(self.test_foo_obj.references, self.test_links)

        test_value = [self.test_link1, self.test_link2]
        self.test_foo_obj.references = test_value
        self.assertEqual(
            sorted(self.test_foo_obj.references, key=lambda x: x.name),
            sorted(test_value, key=lambda x: x.name)
        )

    def test_references_application_test(self):
        """testing an example of ReferenceMixin usage
        """
        class GreatEntity(SimpleEntity, ReferenceMixin):
            __tablename__ = 'GreatEntities'
            __mapper_args__ = {
                "polymorphic_identity": "GreatEntity"
            }
            ge_id = Column('id', Integer, ForeignKey('SimpleEntities.id'),
                           primary_key=True)

        my_ge = GreatEntity(name="Test")
        # we should have a references attribute right now
        var = my_ge.references

        image_link_type = Type(
            name='Image',
            code='image',
            target_entity_type="Link"
        )
        new_link = Link(name="NewTestLink", full_path="nopath",
                        filename="nofilename", type=image_link_type)

        test_value = [new_link]

        my_ge.references = test_value

        self.assertEqual(my_ge.references, test_value)



"""Vacation.user can now be nullable

Revision ID: 130a7697cd79
Revises: 57a5949c7f29
Create Date: 2013-08-02 19:58:59.638085

"""

# revision identifiers, used by Alembic.
revision = '130a7697cd79'
down_revision = '57a5949c7f29'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        'Vacations', 'user_id', existing_type=sa.INTEGER(), nullable=True
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        'Vacations', 'user_id', existing_type=sa.INTEGER(), nullable=False
    )
    ### end Alembic commands ###



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import datetime
import logging

from sqlalchemy import (Table, Column, String, Integer, ForeignKey, Interval,
                        DateTime, PickleType, Float, Enum)
from sqlalchemy.exc import UnboundExecutionError
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import synonym, relationship, validates

from stalker import defaults
from stalker.db.session import DBSession
from stalker.db.declarative import Base
from stalker.log import logging_level
from stalker.models import make_plural

logger = logging.getLogger(__name__)
logger.setLevel(logging_level)


def create_secondary_table(
        primary_cls_name,
        secondary_cls_name,
        primary_cls_table_name,
        secondary_cls_table_name,
        secondary_table_name=None):
    """creates any secondary table
    """
    plural_secondary_cls_name = make_plural(secondary_cls_name)

    # use the given class_name and the class_table
    if not secondary_table_name:
        secondary_table_name = '%s_%s' % (primary_cls_name,
                                          plural_secondary_cls_name)

    # check if the table is already defined
    if secondary_table_name not in Base.metadata:
        secondary_table = Table(
            secondary_table_name, Base.metadata,
            Column(
                '%s_id' % primary_cls_name.lower(),
                Integer,
                ForeignKey('%s.id' % primary_cls_table_name),
                primary_key=True,
            ),

            Column(
                '%s_id' % secondary_cls_name.lower(),
                Integer,
                ForeignKey('%s.id' % secondary_cls_table_name),
                primary_key=True,
            )
        )
    else:
        secondary_table = Base.metadata.tables[secondary_table_name]

    return secondary_table


class TargetEntityTypeMixin(object):
    """Adds target_entity_type attribute to mixed in class.

    :param target_entity_type: The target entity type which this class is
      designed for. Should be a class or a class name.

      For example::

        from stalker import SimpleEntity, TargetEntityTypeMixin, Project

        class A(SimpleEntity, TargetEntityTypeMixin):
            __tablename__ = "As"
            __mapper_args__ = {"polymorphic_identity": "A"}

            def __init__(self, **kwargs):
                super(A, self).__init__(**kwargs)
                TargetEntityTypeMixin.__init__(self, **kwargs)

        a_obj = A(target_entity_type=Project)

      The ``a_obj`` will only be accepted by
      :class:`.Project` instances. You can not assign it to any other class
      which accepts a :class:`.Type` instance.

    To control the mixed-in class behaviour add these class variables to the
    mixed in class:

      __nullable_target__ : controls if the target_entity_type can be
                            nullable or not. Default is False.

      __unique_target__ : controls if the target_entity_type should be
                          unique, so there is only one object for one type.
                          Default is False.
    """

    __nullable_target__ = False
    __unique_target__ = False

    @declared_attr
    def _target_entity_type(cls):
        return Column(
            "target_entity_type",
            String(128),
            nullable=cls.__nullable_target__,
            unique=cls.__unique_target__
        )

    def __init__(self, target_entity_type=None, **kwargs):
        self._target_entity_type = \
            self._validate_target_entity_type(target_entity_type)

    def _validate_target_entity_type(self, target_entity_type_in):
        """validates the given target_entity_type value
        """
        # it can not be None
        if target_entity_type_in is None:
            raise TypeError(
                "%s.target_entity_type can not be None" %
                self.__class__.__name__
            )

        if str(target_entity_type_in) == "":
            raise ValueError(
                "%s.target_entity_type can not be empty" %
                self.__class__.__name__
            )

        # check if it is a class
        if isinstance(target_entity_type_in, type):
            target_entity_type_in = target_entity_type_in.__name__

        return str(target_entity_type_in)

    def _target_entity_type_getter(self):
        return self._target_entity_type

    @declared_attr
    def target_entity_type(cls):
        return synonym(
            "_target_entity_type",
            descriptor=property(
                fget=cls._target_entity_type_getter,
                doc="""The entity type which this object is valid for.

                Usually it is set to the TargetClass directly.
                """
            )
        )


class StatusMixin(object):
    """Makes the mixed in object statusable.

    This mixin adds status and status_list attributes to the mixed in class.
    Any object that needs a status and a corresponding status list can include
    this mixin.

    When mixed with a class which don't have an __init__ method, the mixin
    supplies one, and in this case the parameters below must be defined.

    :param status_list: this attribute holds a status list object, which shows
      the possible statuses that this entity could be in. This attribute can
      not be empty or None. Giving a StatusList object, the
      StatusList.target_entity_type should match the current class.

      .. versionadded:: 0.1.2.a4

        The status_list argument now can be skipped or can be None if there
        is an active database connection and there is a suitable
        :class:`.StatusList` instance in the database whom
        :attr:`.StatusList.target_entity_type` attribute is set to the current
        mixed-in class name.

    :param status: It is a :class:`.Status` instance which shows the current
      status of the statusable object. Integer values are also accepted, which
      shows the index of the desired status in the ``status_list`` attribute of
      the current statusable object. If a :class:`.Status` instance is
      supplied, it should also be present in the ``status_list`` attribute. If
      set to None then the first :class:`.Status` instance in the
      ``status_list`` will be used.

      .. versionadded:: 0.2.0

        Status attribute as Status instance:

        It is now possible to set the status of the instance by a
        :class:`.Status` instance directly. And the :attr:`.StatusMixin.status`
        will return a proper :class:`.Status` instance.
    """

    def __init__(self, status=None, status_list=None, **kwargs):
        self.status_list = status_list
        self.status = status

    @declared_attr
    def status_id(cls):
        return Column(
            'status_id',
            Integer,
            ForeignKey('Statuses.id'),
            nullable=False
            # This is set to nullable=True but it is impossible to set the
            # status to None by using this Declarative approach.
            #
            # This is done in that way cause SQLAlchemy was flushing the data
            # (AutoFlush) preliminarily while checking if the given Status was
            # in the related StatusList, and it was complaining about the
            # status can not be null
        )

    @declared_attr
    def status(cls):
        return relationship(
            'Status',
            primaryjoin= "%s.status_id==Status.status_id" % cls.__name__,
            doc="""The current status of the object.

            It is a :class:`.Status` instance which
            is one of the Statuses stored in the ``status_list`` attribute
            of this object.
            """
        )

    @declared_attr
    def status_list_id(cls):
        return Column(
            'status_list_id',
            Integer,
            ForeignKey('StatusLists.id'),
            nullable=False
        )

    @declared_attr
    def status_list(cls):
        return relationship(
            "StatusList",
            primaryjoin= \
                "%s.status_list_id==StatusList.status_list_id" %
                cls.__name__,
        )

    @validates("status_list")
    def _validate_status_list(self, key, status_list):
        """validates the given status_list_in value
        """
        from stalker.models.status import StatusList

        if status_list is None:
            # check if there is a db setup and try to get the appropriate
            # StatusList from the database

            # disable autoflush to prevent premature class initialization
            with DBSession.no_autoflush:
                try:
                    # try to get a StatusList with the target_entity_type is
                    # matching the class name
                    status_list = StatusList.query\
                        .filter_by(target_entity_type=self.__class__.__name__)\
                        .first()
                except UnboundExecutionError:
                    # it is not mapped just skip it
                    pass

        # if it is still None
        if status_list is None:
            # there is no db so raise an error because there is no way
            # to get an appropriate StatusList
            raise TypeError(
                "%s instances can not be initialized without a "
                "stalker.models.status.StatusList instance, please pass a "
                "suitable StatusList (StatusList.target_entity_type=%s) "
                "with the 'status_list' argument" %
                (self.__class__.__name__, self.__class__.__name__)
            )
        else:
            # it is not an instance of status_list
            if not isinstance(status_list, StatusList):
                raise TypeError(
                    "%s.status_list should be an instance of "
                    "stalker.models.status.StatusList not %s" %
                    (self.__class__.__name__, status_list.__class__.__name__)
                )

            # check if the entity_type matches to the
            # StatusList.target_entity_type
            if self.__class__.__name__ != status_list.target_entity_type:
                raise TypeError(
                    "the given StatusLists' target_entity_type is %s, "
                    "whereas the entity_type of this object is %s" %
                    (status_list.target_entity_type, self.__class__.__name__)
                )

        return status_list

    @validates('status')
    def _validate_status(self, key, status):
        """validates the given status value
        """
        from stalker.models.status import Status, StatusList

        with DBSession.no_autoflush:
            if not isinstance(self.status_list, StatusList):
                raise TypeError(
                    "Please set the %s.status_list attribute first" %
                    self.__class__.__name__
                )

        # it is set to None
        if status is None:
            with DBSession.no_autoflush:
                status = self.status_list.statuses[0]

        # it is not an instance of status or int
        if not isinstance(status, (Status, int)):
            raise TypeError(
                "%(class)s.status must be an instance of "
                "stalker.models.status.Status or an integer showing the index "
                "of the Status object in the %(class)s.status_list, not "
                "%(status)s" %
                {
                    'class': self.__class__.__name__,
                    'status': status.__class__.__name__
                }
            )

        if isinstance(status, int):
            # if it is not in the correct range:
            if status < 0:
                raise ValueError(
                    "%s.status must be a non-negative integer" %
                    self.__class__.__name__
                )

            if status >= len(self.status_list.statuses):
                raise ValueError(
                    "%s.status can not be bigger than the length of the "
                    "status_list" % self.__class__.__name__
                )
                # get the status instance out of the status_list instance
            status = self.status_list[status]

        # check if the given status is in the status_list
        # logger.debug('self.status_list: %s' % self.status_list)
        # logger.debug('given status: %s' % status)

        if status not in self.status_list:
            raise ValueError(
                "The given Status instance for %s.status is not in the "
                "%s.status_list, please supply a status from that list." %
                (self.__class__.__name__, self.__class__.__name__)
            )

        return status


class DateRangeMixin(object):
    """Adds date range info to the mixed in class.

    Adds date range information like ``start``, ``end`` and ``duration``. These
    attributes will be used in TaskJuggler. Because ``effort`` is only
    meaningful if there are some ``resources`` this attribute has been left
    special for :class:`.Task` class. The ``length`` has not been implemented
    because of its rare use.

    The preceding order for the attributes is as follows::

      start > end > duration

    So if all of the parameters are given only the ``start`` and the ``end``
    will be used and the ``duration`` will be calculated accordingly. In any
    other conditions the missing parameter will be calculated from the
    following table:

    +-------+-----+----------+----------------------------------------+
    | start | end | duration | DEFAULTS                               |
    +=======+=====+==========+========================================+
    |       |     |          | start = datetime.datetime.now()        |
    |       |     |          |                                        |
    |       |     |          | duration = datetime.timedelta(days=10) |
    |       |     |          |                                        |
    |       |     |          | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |     |          | duration = datetime.timedelta(days=10) |
    |       |     |          |                                        |
    |       |     |          | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |  X  |          | duration = end - start                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |     |    X     | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |  X  |    X     | duration = end - start                 |
    +-------+-----+----------+----------------------------------------+
    |       |  X  |    X     | start = end - duration                 |
    +-------+-----+----------+----------------------------------------+
    |       |  X  |          | duration = datetime.timedelta(days=10) |
    |       |     |          |                                        |
    |       |     |          | start = end - duration                 |
    +-------+-----+----------+----------------------------------------+
    |       |     |    X     | start = datetime.datetime.now()        |
    |       |     |          |                                        |
    |       |     |          | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+

    Only the ``start``, ``end`` will be stored. The ``duration`` attribute is
    the direct difference of the the ``start`` and ``end`` attributes, so there
    is no need to store it. But if will be used in calculation of the start and
    end values.

    The start and end attributes have a ``computed`` companion. Which are the
    return values from TaskJuggler. so for start there is the
    ``computed_start`` and for end there is the ``computed_end`` attributes.
    These values are going to be used in Gantt Charts.

    The date attributes can be managed with timezones. Follow the Python idioms
    shown in the `documentation of datetime`_

    .. _documentation of datetime: http://docs.python.org/library/datetime.html

    :param start: the start date of the entity, should be a datetime.datetime
      instance, the start is the pin point for the date calculation. In
      any condition if the start is available then the value will be
      preserved. If start passes the end the end is also changed
      to a date to keep the timedelta between dates. The default value is
      datetime.datetime.now()

    :type start: :class:`datetime.datetime`

    :param end: the end of the entity, should be a datetime.datetime instance,
      when the start is changed to a date passing the end, then the end is also
      changed to a later date so the timedelta between the dates is kept.

    :type end: :class:`datetime.datetime` or :class:`datetime.timedelta`

    :param duration: The duration of the entity. It is a
      :class:`datetime.timedelta` instance. The default value is read from
      the :class:`.Config` class. See the table above for the initialization
      rules.

    :type duration: :class:`datetime.timedelta`
    """

    def __init__(
            self,
            start=None,
            end=None,
            duration=None,
            **kwargs
    ):
        self._start, self._end, self._duration = \
            self._validate_dates(start, end, duration)

    @declared_attr
    def _end(cls):
        return Column("end", DateTime)

    def _end_getter(self):
        """The date that the entity should be delivered.

        The end can be set to a datetime.timedelta and in this case it will be
        calculated as an offset from the start and converted to
        datetime.datetime again. Setting the start to a date passing the end
        will also set the end, so the timedelta between them is preserved,
        default value is 10 days
        """
        return self._end

    def _end_setter(self, end_in):
        self._start, self._end, self._duration = \
            self._validate_dates(self.start, end_in, self.duration)

    @declared_attr
    def end(cls):
        return synonym(
            "_end",
            descriptor=property(
                cls._end_getter,
                cls._end_setter
            )
        )

    @declared_attr
    def _start(cls):
        return Column("start", DateTime)

    def _start_getter(self):
        """The date that this entity should start.

        Also effects the :attr:`.DateRangeMixin.end` attribute value in certain
        conditions, if the :attr:`.DateRangeMixin.start` is set to a time
        passing the :attr:`.DateRangeMixin.end` it will also offset the
        :attr:`.DateRangeMixin.end` to keep the
        :attr:`.DateRangeMixin.duration` value fixed.
        :attr:`.DateRangeMixin.start` should be an instance of
        class:`datetime.datetime` and the default value is
        :func:`datetime.datetime.now()`
        """
        return self._start

    def _start_setter(self, start_in):
        self._start, self._end, self._duration = \
            self._validate_dates(start_in, self.end, self.duration)

    @declared_attr
    def start(cls):
        return synonym(
            "_start",
            descriptor=property(
                cls._start_getter,
                cls._start_setter,
            )
        )

    @declared_attr
    def _duration(cls):
        return Column('duration', Interval)

    def _duration_getter(self):
        return self._duration

    def _duration_setter(self, duration_in):
        if duration_in is not None:
            if isinstance(duration_in, datetime.timedelta):
                # set the end to None
                # to make it recalculated
                self._start, self._end, self._duration = \
                    self._validate_dates(self.start, None, duration_in)
            else:
                # use the end
                self._start, self._end, self._duration = \
                    self._validate_dates(self.start, self.end, duration_in)
        else:
            self._start, self._end, self._duration = \
                self._validate_dates(self.start, self.end, duration_in)

    @declared_attr
    def duration(self):
        return synonym(
            '_duration',
            descriptor=property(
                self._duration_getter,
                self._duration_setter,
                doc="""Duration of the entity.

                It is a datetime.timedelta instance. Showing the difference of
                the :attr:`.start` and the :attr:`.end`. If edited it changes
                the :attr:`.end` attribute value."""
            )
        )

    def _validate_dates(self, start, end, duration):
        """updates the date values
        """
        # logger.debug('start    : %s' % start)
        # logger.debug('end      : %s' % end)
        # logger.debug('duration : %s' % duration)

        if not isinstance(start, datetime.datetime):
            start = None

        if not isinstance(end, datetime.datetime):
            end = None

        if not isinstance(duration, datetime.timedelta):
            duration = None

        # check start
        if start is None:
            # try to calculate the start from end and duration
            if end is None:
                # set the defaults
                start = datetime.datetime.now()

                if duration is None:
                    # set the defaults
                    duration = defaults.timing_resolution

                end = start + duration
            else:
                if duration is None:
                    duration = defaults.timing_resolution

                # try:
                start = end - duration
                # except OverflowError: # end is datetime.datetime.min
                #     start = end

        # check end
        if end is None:
            if duration is None:
                duration = defaults.timing_resolution

            end = start + duration

        if end < start:
            # check duration
            if duration is None or duration < datetime.timedelta(1):
                duration = datetime.timedelta(1)

            # try:
            end = start + duration
            # except OverflowError: # start is datetime.datetime.max
            #     end = start

        # round the dates to the timing_resolution
        rounded_start = self.round_time(start)
        rounded_end = self.round_time(end)
        rounded_duration = rounded_end - rounded_start

        if rounded_duration < defaults.timing_resolution:
            rounded_duration = defaults.timing_resolution
            rounded_end = rounded_start + rounded_duration

        return rounded_start, rounded_end, rounded_duration

    @declared_attr
    def computed_start(cls):
        return Column('computed_start', DateTime)

    @declared_attr
    def computed_end(cls):
        return Column('computed_end', DateTime)


    @property
    def computed_duration(self):
        """returns the computed_duration as the difference of computed_start
        and computed_end if there are computed_start and computed_end otherwise
        returns None
        """
        return self.computed_end - self.computed_start \
            if self.computed_end and self.computed_start else None

    @classmethod
    def round_time(cls, dt):
        """Round the given datetime object to the defaults.timing_resolution.

        Uses :class:`stalker.defaults.timing_resolution` as the closest number
        of seconds to round to.

        :param dt: datetime.datetime object, defaults to now.

        :type dt: datetime.datetime

        Based on Thierry Husson's answer in `Stackoverflow`_

        _`Stackoverflow` : http://stackoverflow.com/a/10854034/1431079
        """
        # to be compatible with python 2.6 use the following instead of
        # total_seconds()
        timing_resolution = defaults.timing_resolution
        trs = timing_resolution.days * 86400 + timing_resolution.seconds

        # convert to seconds
        epoch = datetime.datetime(1970, 1, 1)
        diff = dt - epoch
        diff_in_seconds = diff.days * 86400 + diff.seconds
        return epoch + datetime.timedelta(
            seconds=(diff_in_seconds + trs * 0.5) // trs * trs
        )

    @property
    def total_seconds(self):
        """returns the duration as seconds
        """
        return self.duration.days * 86400 + self.duration.seconds

    @property
    def computed_total_seconds(self):
        """returns the duration as seconds
        """
        return self.computed_duration.days * 86400 + \
            self.computed_duration.seconds


class ProjectMixin(object):
    """Allows connecting a :class:`.Project` to the mixed in object.

    :param project: A :class:`.Project` instance holding the project which this
      object is related to. It can not be None, or anything other than a
      :class:`.Project` instance.

    :type project: :class:`.Project`
    """

    #    # add this lines for Sphinx
    #    __tablename__ = "ProjectMixins"

    @declared_attr
    def project_id(cls):
        return Column(
            "project_id",
            Integer,
            ForeignKey("Projects.id", use_alter=True, name="project_x_id"),
            #ForeignKey("Projects.id"),
            # cannot use nullable cause a Project object needs
            # insert itself as the project and it needs post_update
            # thus nullable should be True
            #nullable=False,
        )

    @declared_attr
    def project(cls):
        backref = cls.__tablename__.lower()
        doc = """The :class:`.Project` instance that
        this object belongs to.
        """

        return relationship(
            "Project",
            primaryjoin=
            cls.__tablename__ + ".c.project_id==Projects.c.id",
            post_update=True,  # for project itself
            uselist=False,
            backref=backref,
            doc=doc
        )

    def __init__(self,
                 project=None,
                 **kwargs):
        self.project = project

    @validates("project")
    def _validate_project(self, key, project):
        """validates the given project value
        """
        from stalker.models.project import Project
        if project is None:
            raise TypeError(
                "%s.project can not be None it must be an instance of "
                "stalker.models.project.Project" % self.__class__.__name__
            )

        if not isinstance(project, Project):
            raise TypeError(
                "%s.project should be an instance of "
                "stalker.models.project.Project instance not %s" %
                (self.__class__.__name__, project.__class__.__name__)
            )
        return project


class ReferenceMixin(object):
    """Adds reference capabilities to the mixed in class.

    References are :class:`stalker.models.link.Link` instances or anything
    derived from it, which adds information to the attached objects. The aim of
    the References are generally to give more info to direct the evolution of
    the object.

    :param references: A list of :class:`.Link` instances.

    :type references: list of :class:`.Link` instances.
    """
    # add this lines for Sphinx
    #    __tablename__ = "ReferenceMixins"

    def __init__(self,
                 references=None,
                 **kwargs):
        if references is None:
            references = []

        self.references = references

    @declared_attr
    def references(cls):
        # get secondary table
        secondary_table = create_secondary_table(
            cls.__name__,
            'Link',
            cls.__tablename__,
            'Links',
            cls.__name__ + "_References"
        )
        # return the relationship
        return relationship(
            "Link",
            secondary=secondary_table,
            doc="""A list of :class:`.Link` instances given as a reference for
            this entity.
            """
        )

    @validates("references")
    def _validate_references(self, key, reference):
        """validates the given reference
        """
        from stalker.models.link import Link

        # all the elements should be instance of stalker.models.entity.Entity
        if not isinstance(reference, Link):
            raise TypeError(
                "%s.references should be all instances of "
                "stalker.models.link.Link not %s" %
                (self.__class__.__name__, reference.__class__.__name__)
            )
        return reference


class ACLMixin(object):
    """A Mixin for adding ACLs to mixed in class.

    Access control lists or ACLs are used to determine if the given resource
    has the permission to access the given data. It is based on Pyramids
    Authorization system but organized to fit in Stalker style.

    The ACLMixin adds an attribute called ``permissions`` and a
    property called ``__acl__`` to be able to pass the permission data to
    Pyramid framework.
    """

    @declared_attr
    def permissions(cls):
        # get the secondary table
        secondary_table = create_secondary_table(
            cls.__name__, 'Permission', cls.__tablename__, 'Permissions'
        )
        return relationship('Permission', secondary=secondary_table)

    @validates('permissions')
    def _validate_permissions(self, key, permission):
        """validates the given permission value
        """
        from stalker.models.auth import Permission
        if not isinstance(permission, Permission):
            raise TypeError(
                "%s.permissions should be all instances of "
                "stalker.models.auth.Permission not %s" %
                (self.__class__.__name__, permission.__class__.__name__)
            )

        return permission

    @property
    def __acl__(self):
        """Returns Pyramid friendly ACL list composed by the:

          * Permission.access (Ex: 'Allow' or 'Deny')
          * The Mixed in class name and the object name (Ex: 'User:eoyilmaz')
          * The Action and the target class name (Ex: 'Create_Asset')

        Thus a list of tuple is returned as follows::

          __acl__ = [
              ('Allow', 'User:eoyilmaz', 'Create_Asset'),
          ]

        For the last example user eoyilmaz can grant access to views requiring
        'Add_Project' permission.
        """
        return [(perm.access,
                 '%s:%s' % (self.__class__.__name__, self.name),
                 perm.action + '_' + perm.class_name)
                for perm in self.permissions]


class CodeMixin(object):
    """Adds code info to the mixed in class.

    .. versionadded:: 0.2.0

      The code attribute of the SimpleEntity is now introduced as a separate
      mixin. To let it be used by the classes it is really needed.

    The CodeMixin just adds a new field called ``code``. It is a very simple
    attribute and is used for simplifying long names (like Project.name etc.).

    Contrary to previous implementations the code attribute is not formatted in
    anyway, so care needs to be taken if the code attribute is going to be used
    in filesystem as file and directory names.

    :param str code: The code attribute is a string, can not be empty or can
      not be None.
    """

    def __init__(self, code=None, **kwargs):
        logger.debug('code: %s' % code)
        self.code = code

    @declared_attr
    def code(cls):
        return Column(
            'code',
            String(256),
            nullable=False,
            doc="""The code name of this object.

                It accepts strings. Can not be None."""
        )

    @validates('code')
    def _validate_code(self, key, code):
        """validates the given code attribute
        """
        logger.debug('validating code value of: %s' % code)
        if code is None:
            raise TypeError("%s.code cannot be None" % self.__class__.__name__)

        if not isinstance(code, str):
            raise TypeError(
                '%s.code should be a string not %s' %
                (self.__class__.__name__, code.__class__.__name__)
            )

        if code == '':
            raise ValueError(
                '%s.code can not be an empty string' % self.__class__.__name__
            )

        return code


class WorkingHoursMixin(object):
    """Sets working hours for the mixed in class.

    Generally is meaningful for users, departments and studio.

    :param working_hours: A :class:`.WorkingHours` instance showing the working
      hours settings for that project. This data is stored as a PickleType in
      the database.
    """

    def __init__(self, working_hours=None, **kwargs):
        self.working_hours = working_hours

    @declared_attr
    def working_hours(cls):
        return Column(PickleType)

    @validates('working_hours')
    def _validate_working_hours(self, key, wh):
        """validates the given working hours value
        """
        if wh is None:
            # use the default one
            from stalker import WorkingHours
            wh = WorkingHours()

        return wh


class ScheduleMixin(object):
    """Adds schedule info to the mixed in class.

    Adds attributes like schedule_timing, schedule_unit and schedule_model
    attributes to the mixed in class.

    Use the ``__default_schedule_attr_name__`` attribute to customize the
    column names.
    """

    # some default values that can be overridden in Mixed in classes
    __default_schedule_attr_name__ = 'schedule'
    __default_schedule_timing__ = defaults.timing_resolution.seconds / 60
    __default_schedule_unit__ = 'h'
    __default_schedule_models__ = defaults.task_schedule_models

    def __init__(
            self,
            schedule_timing=None,
            schedule_unit=None,
            schedule_model=None,
            schedule_constraint=0,
            **kwargs
    ):
        self.schedule_constraint = schedule_constraint
        self.schedule_model = schedule_model
        self.schedule_timing = schedule_timing
        self.schedule_unit = schedule_unit

    @declared_attr
    def schedule_timing(cls):
        return Column(
            '%s_timing' % cls.__default_schedule_attr_name__,
            Float, nullable=True, default=0,
            doc="""It is the value of the %(attr)s timing. It is a float
            value.

            The timing value can either be as Work Time or Calendar Time
            defined by the %(attr)s_model attribute. So when the %(attr)s_model
            is `duration` then the value of this attribute is in Calendar Time,
            and if the %(attr)s_model is either `length` or `effort` then the
            value is considered as Work Time.
            """ % {'attr': cls.__default_schedule_attr_name__}
        )

    @declared_attr
    def schedule_unit(cls):
        return Column(
            '%s_unit' % cls.__default_schedule_attr_name__,
            Enum(*defaults.datetime_units, name='TimeUnit'),
            nullable=False, default='h',
            doc="""It is the unit of the %(attr)s timing. It is a string
            value. And should be one of 'min', 'h', 'd', 'w', 'm', 'y'.""" %
                {'attr': cls.__default_schedule_attr_name__}
        )

    @declared_attr
    def schedule_model(cls):
        return Column(
            '%s_model' % cls.__default_schedule_attr_name__,
            Enum(*cls.__default_schedule_models__,
                 name='%(class)s%(attr)sModel' % {
                     'class': cls.__name__,
                     'attr': cls.__default_schedule_attr_name__.title()
                 }),
            default=cls.__default_schedule_models__[0], nullable=False,
            doc="""Defines the schedule model which is going to be used by
            **TaskJuggler** while scheduling this Task. It has three possible
            values; **effort**, **duration**, **length**. ``effort`` is the
            default value. Each value causes this task to be scheduled in
            different ways:

            ======== ==========================================================
            effort   If the :attr:`.schedule_model` attribute is set to
                     **"effort"** then the start and end date values are
                     calculated so that a resource should spent this much of
                     work time to complete a Task. For example, a task with
                     :attr:`.schedule_timing` of 4 days, needs 4 working days.
                     So it can take 4 working days to complete the Task, but it
                     doesn't mean that the task duration will be 4 days. If the
                     resource works overtime then the task will be finished
                     before 4 days or if the resource will not be available
                     (due to a vacation) then the task duration can be much
                     more.

            duration The duration of the task will exactly be equal to
                     :attr:`.schedule_timing` regardless of the resource
                     availability. So the difference between :attr:`.start`
                     and :attr:`.end` attribute values are equal to
                     :attr:`.schedule_timing`. Essentially making the task
                     duration in calendar days instead of working days.

            length   In this model the duration of the task will exactly be
                     equal to the given length value in working days regardless
                     of the resource availability. So a task with the
                     :attr:`.schedule_timing` is set to 4 days will be
                     completed in 4 working days. But again it will not be
                     always 4 calendar days due to the weekends or non working
                     days.
            ======== ==========================================================
            """
        )

    @declared_attr
    def schedule_constraint(cls):
        return Column(
            '%s_constraint' % cls.__default_schedule_attr_name__,
            Integer,
            default=0,
            nullable=False,
            doc="""An integer number showing the constraint schema for this
            task.

            Possible values are:

             ===== ===============
               0   Constrain None
               1   Constrain Start
               2   Constrain End
               3   Constrain Both
             ===== ===============

            For convenience use **stalker.models.task.CONSTRAIN_NONE**,
            **stalker.models.task.CONSTRAIN_START**,
            **stalker.models.task.CONSTRAIN_END**,
            **stalker.models.task.CONSTRAIN_BOTH**.

            This value is going to be used to constrain the start and end date
            values of this task. So if you want to pin the start of a task to a
            certain date. Set its :attr:`.schedule_constraint` value to
            **CONSTRAIN_START**. When the task is scheduled by **TaskJuggler**
            the start date will be pinned to the :attr:`start` attribute of
            this task.

            And if both of the date values (start and end) wanted to be pinned
            to certain dates (making the task effectively a ``duration`` task)
            set the desired :attr:`start` and :attr:`end` and then set the
            :attr:`schedule_constraint` to **CONSTRAIN_BOTH**.
            """
        )

    @validates('schedule_constraint')
    def _validate_schedule_constraint(self, key, schedule_constraint):
        """validates the given schedule_constraint value
        """
        if not schedule_constraint:
            schedule_constraint = 0

        if not isinstance(schedule_constraint, int):
            raise TypeError(
                '%(class)s.%(attr)s_constraint should be an integer '
                'between 0 and 3, not %(constraint_class)s' %
                {
                    'class': self.__class__.__name__,
                    'attr': self.__default_schedule_attr_name__,
                    'constraint_class':
                    schedule_constraint.__class__.__name__
                }
            )

        schedule_constraint = max(schedule_constraint, 0)
        schedule_constraint = min(schedule_constraint, 3)

        return schedule_constraint

    @validates('schedule_model')
    def _validate_schedule_model(self, key, schedule_model):
        """validates the given schedule_model value
        """
        if not schedule_model:
            schedule_model = self.__default_schedule_models__[0]

        error_message = \
            '%(class)s.%(attr)s_model should be one of %(defaults)s, not ' \
            '%(model_class)s' % {
                'class': self.__class__.__name__,
                'attr': self.__default_schedule_attr_name__,
                'defaults': self.__default_schedule_models__,
                'model_class': schedule_model.__class__.__name__
            }

        if not isinstance(schedule_model, str):
            raise TypeError(error_message)

        if schedule_model not in self.__default_schedule_models__:
            raise ValueError(error_message)

        return schedule_model

    @validates('schedule_unit')
    def _validate_schedule_unit(self, key, schedule_unit):
        """validates the given schedule_unit
        """
        if schedule_unit is None:
            schedule_unit = self.__default_schedule_unit__

        if not isinstance(schedule_unit, str):
            raise TypeError(
                '%(class)s.%(attr)s_unit should be a string value one of '
                '%(defaults)s showing the unit of the %(attr)s timing of this '
                '%(class)s, not %(unit_class)s' % {
                    'class': self.__class__.__name__,
                    'attr': self.__default_schedule_attr_name__,
                    'defaults': defaults.datetime_units,
                    'unit_class': schedule_unit.__class__.__name__
                }
            )

        if schedule_unit not in defaults.datetime_units:
            raise ValueError(
                '%(class)s.%(attr)s_unit should be a string value one of '
                '%(defaults)s showing the unit of the %(attr)s timing of '
                'this %(class)s, not %(unit_class)s' % {
                    'class': self.__class__.__name__,
                    'attr': self.__default_schedule_attr_name__,
                    'defaults': defaults.datetime_units,
                    'unit_class': schedule_unit.__class__.__name__
                }
            )

        return schedule_unit

    @validates('schedule_timing')
    def _validate_schedule_timing(self, key, schedule_timing):
        """validates the given schedule_timing
        """
        if schedule_timing is None:
            schedule_timing = self.__default_schedule_timing__
            self.schedule_unit = self.__default_schedule_unit__

        if not isinstance(schedule_timing, (int, float)):
            raise TypeError(
                '%(class)s.%(attr)s_timing should be an integer or float '
                'number showing the value of the %(attr)s timing of this '
                '%(class)s, not %(timing_class)s' % {
                    'class': self.__class__.__name__,
                    'attr': self.__default_schedule_attr_name__,
                    'timing_class': schedule_timing.__class__.__name__
                }
            )

        return schedule_timing

    @classmethod
    def least_meaningful_time_unit(cls, seconds, as_work_time=True):
        """returns the least meaningful timing unit that corresponds to the
        given seconds. So if:

          as_work_time == True
              seconds % (1 years work time as seconds) == 0 --> 'y' else:
              seconds % (1 month work time as seconds) == 0 --> 'm' else:
              seconds % (1 week work time as seconds) == 0 --> 'w' else:
              seconds % (1 day work time as seconds) == 0 --> 'd' else:
              seconds % (1 hour work time as seconds) == 0 --> 'h' else:
              seconds % (1 minutes work time as seconds) == 0 --> 'min' else:
              raise RuntimeError
          as_work_time == False
              seconds % (1 years as seconds) == 0 --> 'y' else:
              seconds % (1 month as seconds) == 0 --> 'm' else:
              seconds % (1 week as seconds) == 0 --> 'w' else:
              seconds % (1 day as seconds) == 0 --> 'd' else:
              seconds % (1 hour as seconds) == 0 --> 'h' else:
              seconds % (1 minutes as seconds) == 0 --> 'min' else:
              raise RuntimeError

        :param int seconds: An integer showing the total seconds to be
          converted.
        :param bool as_work_time: Should the input be considered as work time
          or calendar time.
        :returns int, string: Returns one integer and one string, showing the
          timing value and the unit.
        """
        minutes = 60
        hour = 3600
        day = 86400
        week = 604800
        month = 2419200
        year = 31536000

        day_wt = defaults.daily_working_hours * 3600
        week_wt = defaults.weekly_working_days * day_wt
        month_wt = 4 * week_wt
        year_wt = int(defaults.yearly_working_days) * day_wt

        if as_work_time:
            logger.debug('calculating in work time')
            if seconds % year_wt == 0:
                return seconds // year_wt, 'y'
            elif seconds % month_wt == 0:
                return seconds // month_wt, 'm'
            elif seconds % week_wt == 0:
                return seconds // week_wt, 'w'
            elif seconds % day_wt == 0:
                return seconds // day_wt, 'd'
        else:
            logger.debug('calculating in calendar time')
            if seconds % year == 0:
                return seconds // year, 'y'
            elif seconds % month == 0:
                return seconds // month, 'm'
            elif seconds % week == 0:
                return seconds // week, 'w'
            elif seconds % day == 0:
                return seconds // day, 'd'

        # in either case
        if seconds % hour == 0:
            return seconds // hour, 'h'

        # at this point we understand that it has a residual of less then one
        # minute so return in minutes
        return seconds // minutes, 'min'

    @classmethod
    def to_seconds(cls, timing, unit, model):
        """converts the schedule values to seconds, depending on to the
        schedule_model the value will differ. So if the schedule_model is
        'effort' or 'length' then the schedule_time and schedule_unit values
        are interpreted as work time, if the schedule_model is 'duration' then
        the schedule_time and schedule_unit values are considered as calendar
        time.
        """
        if not unit:
            return None

        lut = {
            'min': 60,
            'h': 3600,
            'd': 86400,
            'w': 604800,
            'm': 2419200,
            'y': 31536000
        }

        if model in ['effort', 'length']:
            day_wt = defaults.daily_working_hours * 3600
            week_wt = defaults.weekly_working_days * day_wt
            month_wt = 4 * week_wt
            year_wt = int(defaults.yearly_working_days) * day_wt

            lut = {
                'min': 60,
                'h': 3600,
                'd': day_wt,
                'w': week_wt,
                'm': month_wt,
                'y': year_wt
            }

        return timing * lut[unit]

    @property
    def schedule_seconds(self):
        """Returns the schedule values as seconds, depending on to the
        schedule_model the value will differ. So if the schedule_model is
        'effort' or 'length' then the schedule_time and schedule_unit values
        are interpreted as work time, if the schedule_model is 'duration' then
        the schedule_time and schedule_unit values are considered as calendar
        time.
        """
        return self.to_seconds(
            self.schedule_timing, self.schedule_unit, self.schedule_model
        )



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest
from stalker import Type, Repository, Status, StatusList, Project
from stalker.models.wiki import Page


class PageTester(unittest.TestCase):
    """Tests stalker.wiki.Page class
    """

    def setUp(self):
        """setting up the test
        """
        # create a repository
        self.repository_type = Type(
            name="Test Repository Type",
            code='test_repo',
            target_entity_type=Repository
        )

        self.test_repository = Repository(
            name="Test Repository",
            type=self.repository_type,
        )

        # statuses
        self.status1 = Status(name="Status1", code="STS1")
        self.status2 = Status(name="Status2", code="STS2")
        self.status3 = Status(name="Status3", code="STS3")

        # project status list
        self.project_status_list = StatusList(
            name="Project Status List",
            statuses=[
                self.status1,
                self.status2,
                self.status3,
            ],
            target_entity_type=Project
        )

        # project type
        self.test_project_type = Type(
            name="Test Project Type",
            code='testproj',
            target_entity_type=Project,
        )

        # create projects
        self.test_project1 = Project(
            name="Test Project 1",
            code='tp1',
            type=self.test_project_type,
            status_list=self.project_status_list,
            repository=self.test_repository,
        )

        self.kwargs = {
            'title': 'Test Page Title',
            'content': 'Test content',
            'project': self.test_project1
        }

        self.test_page = Page(**self.kwargs)

    def test_title_argument_is_skipped(self):
        """testing if a ValueError will be raised when the title argument is
        skipped
        """
        self.kwargs.pop('title')
        with self.assertRaises(ValueError) as cm:
            Page(**self.kwargs)

        self.assertEqual(
            'Page.title can not be empty',
            str(cm.exception)
        )

    def test_title_argument_is_None(self):
        """testing if a TypeError will be raised when the title argument is
        None
        """
        self.kwargs['title'] = None
        with self.assertRaises(TypeError) as cm:
            Page(**self.kwargs)

        self.assertEqual(
            'Page.title should be a string, not NoneType',
            str(cm.exception)
        )

    def test_title_attribute_is_set_to_None(self):
        """testing if a TypeError will be raised when the title attribute is
        set to None
        """
        with self.assertRaises(TypeError) as cm:
            self.test_page.title = None

        self.assertEqual(
            'Page.title should be a string, not NoneType',
            str(cm.exception)
        )

    def test_title_argument_is_an_empty_string(self):
        """testing if a ValueError will be raised when the title argument is
        an empty string
        """
        self.kwargs['title'] = ''
        with self.assertRaises(ValueError) as cm:
            Page(**self.kwargs)

        self.assertEqual(
            'Page.title can not be empty',
            str(cm.exception)
        )

    def test_title_attribute_is_set_to_empty_string(self):
        """testing if a ValueError will be raised when the title attribute is
        set to empty string
        """
        with self.assertRaises(ValueError) as cm:
            self.test_page.title = ''

        self.assertEqual(
            'Page.title can not be empty',
            str(cm.exception)
        )

    def test_title_argument_is_not_a_string(self):
        """testing if a TypeError will be raised when the title argument is not
        a string
        """
        self.kwargs['title'] = 2165
        with self.assertRaises(TypeError) as cm:
            Page(**self.kwargs)

        self.assertEqual(
            'Page.title should be a string, not int',
            str(cm.exception)
        )

    def test_title_attribute_is_not_a_string(self):
        """testing if a TypeError will be raised when the title is set to a
        value other than a string
        """
        with self.assertRaises(TypeError) as cm:
            self.test_page.title = 2135

        self.assertEqual(
            'Page.title should be a string, not int',
            str(cm.exception)
        )

    def test_title_argument_is_working_properly(self):
        """testing if the title argument value is correctly passed to title
        attribute
        """
        self.assertEqual(
            self.kwargs['title'],
            self.test_page.title
        )

    def test_title_attribute_is_working_properly(self):
        """testing if the title attribute is working properly
        """
        test_value = 'Test Title 2'
        self.test_page.title = test_value
        self.assertEqual(
            test_value,
            self.test_page.title
        )

    def test_content_argument_skipped(self):
        """testing if the content attribute value will be an empty string if
        the content argument is skipped
        """
        self.kwargs.pop('content')
        new_page = Page(**self.kwargs)
        self.assertEqual(new_page.content, '')

    def test_content_argument_is_None(self):
        """testing if the content attribute value will be an empty string if
        the content argument is None
        """
        self.kwargs['content'] = None
        new_page = Page(**self.kwargs)
        self.assertEqual(new_page.content, '')

    def test_content_attribute_is_set_to_None(self):
        """testing if the content attribute value will be an empty string if
        the content attribute is set to None
        """
        self.assertNotEqual('', self.test_page.content)
        self.test_page.content = None
        self.assertEqual('', self.test_page.content)

    def test_content_argument_is_empty_string(self):
        """testing if the content attribute value will be an empty string if
        the content argument is an empty string
        """
        self.kwargs['content'] = ''
        new_page = Page(**self.kwargs)
        self.assertEqual(new_page.content, '')

    def test_content_attribute_is_set_to_an_empty_string(self):
        """testing if the content attribute can be set to an empty string
        """
        self.test_page.content = ''
        self.assertEqual('', self.test_page.content)

    def test_content_argument_is_not_a_string(self):
        """testing if a TypeError will be raised when the content argument is
        not a string
        """
        self.kwargs['content'] = 1234
        with self.assertRaises(TypeError) as cm:
            Page(**self.kwargs)

        self.assertEqual(
            'Page.content should be a string, not int',
            str(cm.exception)
        )

    def test_content_attribute_is_set_to_a_value_other_than_a_string(self):
        """testing if a TypeError will be raised when the content attribute is
        set to a value other than a string
        """
        with self.assertRaises(TypeError) as cm:
            self.test_page.content = ['not', 'a', 'string']
        self.assertEqual(
            'Page.content should be a string, not list',
            str(cm.exception)
        )

    def test_content_argument_is_working_properly(self):
        """testing if content argument value is correctly passed to the content
        attribute
        """
        self.assertEqual(
            self.kwargs['content'],
            self.test_page.content
        )

    def test_content_attribute_is_working_properly(self):
        """testing if the content attribute value can be correctly set
        """
        test_value = 'This is a test content'
        self.assertNotEqual(
            test_value,
            self.test_page.content
        )
        self.test_page.content = test_value
        self.assertEqual(
            test_value,
            self.test_page.content
        )



"""Versions are not mixed with StatusMixin anymore

Revision ID: a6598cde6b
Revises: 275bdc106fd5
Create Date: 2013-10-25 17:35:42.953516

"""

# revision identifiers, used by Alembic.
revision = 'a6598cde6b'
down_revision = '275bdc106fd5'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.drop_column('Versions', 'status_list_id')
    op.drop_column('Versions', 'status_id')


def downgrade():
    op.add_column(
        'Versions', sa.Column('status_id', sa.INTEGER(), nullable=False)
    )
    op.add_column(
        'Versions', sa.Column('status_list_id', sa.INTEGER(), nullable=False)
    )



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest
from stalker import Entity, Status


class StatusTest(unittest.TestCase):
    """tests the stalker.models.status.Status class
    """

    def setUp(self):
        """setup the test
        """
        self.kwargs = {
            'name': 'Complete',
            'description': 'use this when the object is complete',
            'code': 'CMPLT',
        }

        # create an entity object with same kwargs for __eq__ and __ne__ tests
        # (it should return False for __eq__ and True for __ne__ for same
        # kwargs)
        self.entity1 = Entity(**self.kwargs)

    def test___auto_name__class_attribute_is_set_to_False(self):
        """testing if the __auto_name__ class attribute is set to False for
        Status class
        """
        self.assertFalse(Status.__auto_name__)

    def test_equality(self):
        """testing equality of two statuses
        """
        status1 = Status(**self.kwargs)
        status2 = Status(**self.kwargs)

        self.kwargs["name"] = "Work In Progress"
        self.kwargs["description"] = "use this when the object is still in \
        progress"
        self.kwargs["code"] = "WIP"

        status3 = Status(**self.kwargs)

        self.assertTrue(status1 == status2)
        self.assertFalse(status1 == status3)
        self.assertFalse(status1 == self.entity1)

    def test_status_and_string_equality_in_status_name(self):
        """testing a status can be compared with a string and returns True if
        the string matches the name and vice versa
        """
        a_status = Status(**self.kwargs)
        self.assertTrue(a_status == self.kwargs["name"])
        self.assertTrue(a_status == self.kwargs["name"].lower())
        self.assertTrue(a_status == self.kwargs["name"].upper())
        self.assertFalse(a_status == "another name")

    def test_status_and_string_equality_in_status_code(self):
        """testing a status can be compared with a string and returns True if
        the string matches the code and vice versa
        """
        a_status = Status(**self.kwargs)
        self.assertTrue(a_status == self.kwargs["code"])
        self.assertTrue(a_status == self.kwargs["code"].lower())
        self.assertTrue(a_status == self.kwargs["code"].upper())

    def test_inequality(self):
        """testing inequality of two statuses
        """
        status1 = Status(**self.kwargs)
        status2 = Status(**self.kwargs)

        self.kwargs["name"] = "Work In Progress"
        self.kwargs["description"] = "use this when the object is still in \
        progress"
        self.kwargs["code"] = "WIP"

        status3 = Status(**self.kwargs)

        self.assertFalse(status1 != status2)
        self.assertTrue(status1 != status3)
        self.assertTrue(status1 != self.entity1)

    def test_status_and_string_inequality_in_status_name(self):
        """testing a status can be compared with a string and returns False if
        the string matches the name and vice versa
        """
        a_status = Status(**self.kwargs)
        self.assertFalse(a_status != self.kwargs["name"])
        self.assertFalse(a_status != self.kwargs["name"].lower())
        self.assertFalse(a_status != self.kwargs["name"].upper())
        self.assertTrue(a_status != "another name")

    def test_status_and_string_inequality_in_status_code(self):
        """testing a status can be compared with a string and returns False if
        the string matches the code and vice versa
        """
        a_status = Status(**self.kwargs)
        self.assertFalse(a_status != self.kwargs["code"])
        self.assertFalse(a_status != self.kwargs["code"].lower())
        self.assertFalse(a_status != self.kwargs["code"].upper())
